#include <stdint.h>
#include "graph.h"

																	/* z VRAM */
static uint8_t						m_vram[(SCREEN_WIDTH/8)*SCREEN_HEIGHT] ;

static const uint8_t				m_shift[]={
										1 << 0,
										1 << 1,
										1 << 2,
										1 << 3,
										1 << 4,
										1 << 5,
										1 << 6,
										1 << 7
									} ;


/* -------------------------------------------------------------------------------- */
/*   _ݒ                                                                       */
/*                                                                                  */
/*   IN : x = W                                                                */
/*        y = W                                                                */
/*   OUT: Ȃ                                                                      */
/* -------------------------------------------------------------------------------- */
void graph_setPixel(uint8_t x, uint8_t y)
{
	m_vram[(y*(SCREEN_WIDTH/8))+(x/8)] |= m_shift[x&7] ;
}


/* -------------------------------------------------------------------------------- */
/*   _                                                                       */
/*                                                                                  */
/*   IN : x = W                                                                */
/*        y = W                                                                */
/*   OUT: Ȃ                                                                      */
/* -------------------------------------------------------------------------------- */
void graph_clearPixel(uint8_t x, uint8_t y)
{
	m_vram[(y*(SCREEN_WIDTH/8))+(x/8)] &= ~m_shift[x&7] ;
}


/* -------------------------------------------------------------------------------- */
/*   z VRAM NA                                                             */
/*                                                                                  */
/*   IN : Ȃ                                                                      */
/*   OUT: Ȃ                                                                      */
/* -------------------------------------------------------------------------------- */
void graph_clear( void ) __naked
{
	__asm
		LD		IY, #0
		ADD		IY, SP

		LD		A, #0
		LD		D, A
		LD		E, A
		LD		B, #(105 / 3)
		LD		HL, #(_m_vram+(168/8)*105)
		LD		SP, HL

clear_100:
		.rept	(168/8)*3
		PUSH	DE
		.endm

		DJNZ	clear_100

		LD		SP, IY
		RET
	__endasm ;
}


/* -------------------------------------------------------------------------------- */
/*   z VRAM  VRAM փRs[                                                     */
/*                                                                                  */
/*   IN : Ȃ                                                                      */
/*   OUT: Ȃ                                                                      */
/* -------------------------------------------------------------------------------- */
void graph_copy( void ) __naked
{
	__asm
		LD		A, #(105 / 5)

		LD		B, #0
		LD		HL, #_m_vram
		LD		DE, #(0xc000 + (16*(240/8))+(36/8))

		AND		A							; CY = 0

copy_100:
		.rept	5
		LD		C, #(168/8)
		LDIR

		LD		C, #(168/8)					; HL -= (168 / 8)
		SBC		HL, BC						;

		EX		DE, HL						; DE += ((240 - 168) / 8)
		LD		C, #((240-168)/8)			;
		ADD		HL, BC						;
		EX		DE, HL						;

		LD		C, #(168/8)
		LDIR

		EX		DE, HL						; DE += ((240 - 168) / 8)
		LD		C, #((240-168)/8)			;
		ADD		HL, BC						;
		EX		DE, HL						;

		.endm

		DEC		A
		JR		NZ, copy_100

		RET
	__endasm ;
}


/* -------------------------------------------------------------------------------- */
/*   `                                                                       */
/*                                                                                  */
/*   IN : x0 = JnW                                                           */
/*        y0 = JnW                                                           */
/*        x1 = IW                                                           */
/*        y1 = IW                                                           */
/*   OUT: Ȃ                                                                      */
/* -------------------------------------------------------------------------------- */
void graph_lineC(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
	int16_t		add ;
	uint8_t		count, data, *p_vram ;
	uint16_t	value, width, height ;


	if( x0 > x1 ) {
		data = x0 ;
		x0   = x1 ;
		x1   = data ;

		data = y0 ;
		y0   = y1 ;
		y1   = data ;
	}
	width  = (x1 - x0) ;

	height = (y1 - y0) ;
	add    = SCREEN_WIDTH / 8 ;
	if( y0 > y1 ) {
		height = -height ;
		add    = -SCREEN_WIDTH / 8 ;
	}

	p_vram = (uint8_t *)&m_vram[(y0*(SCREEN_WIDTH/8))+(x0/8)] ;
	data   = m_shift[x0&7] ;

	if( width >= height ) {		/*  */

		value = width ;
		count = width + 1 ;
		do {
			*p_vram |= data ;
			data     = (data << 1) | (data >> 7) ;
			p_vram  += (data & 1) ;

			value -= height ;
			if( (int16_t )value < 0 ) {
				value  += width ;
				p_vram += add ;
			}
		} while( --count ) ;

	} else {					/*  */

		value = height ;
		count = height + 1 ;
		do {
			*p_vram |= data ;
			p_vram  += add ;

			value -= width ;
			if( (int16_t )value < height ) {
				value  += height ;
				data    = (data << 1) | (data >> 7) ;
				p_vram += (data & 1) ;
			}
		} while( --count ) ;

	}
}


/* -------------------------------------------------------------------------------- */
/*   `                                                                       */
/*                                                                                  */
/*   IN : x0 = JnW                                                           */
/*        y0 = JnW                                                           */
/*        x1 = IW                                                           */
/*        y1 = IW                                                           */
/*   OUT: Ȃ                                                                      */
/* -------------------------------------------------------------------------------- */
void graph_lineASM(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1) __naked
{
	x0 = x0 ;
	y0 = y0 ;
	x1 = x1 ;
	y1 = y1 ;

	__asm
		PUSH	IX
		LD		IY, #0
		ADD		IY, SP

		LD		D, 4(IY)			; x0
		LD		E, 5(IY)			; y0
		LD		H, 6(IY)			; x1
		LD		L, 7(IY)			; y1

graph_line_000:
		; if( x0 <= x1 ) goto graph_line_010
		LD		A, H
		CP		D
		JR		NC, graph_line_010

		EX		DE, HL				; x0:y0 <=> x1:y1

graph_line_010:
		LD		A, H				; B = x1 - x0
		SUB		A, D
		LD		B, A

		LD		A, L				; C = y1 - y0
		SUB		A, E

		LD		IX, #21
		JR		NC, graph_line_020

		NEG
		LD		IX, #-21

graph_line_020:
		LD		C, A
		PUSH	BC

graph_line_100:
		LD		A, D				; HL = &m_vram[(y0*21)+(x0/8)]
		RRCA
		RRCA
		RRCA
		AND		A, #0x1f
		LD		B, #0
		LD		C, E
		LD		H, B
		LD		L, C
		ADD		HL, HL
		ADD		HL, HL
		ADD		HL, BC
		ADD		HL, HL
		ADD		HL, HL
		ADD		HL, BC
		LD		C, A
		ADD		HL, BC
		LD		BC, #_m_vram
		ADD		HL, BC

		LD		A, D				; A = x0 & 7
		AND		#7

		POP		DE

		PUSH	IX					; SP <=> IX
		PUSH	HL					; IX <=> HL
		POP		IX					;
		POP		HL					;
		LD		SP, HL				;

		LD		HL, #_m_shift
		LD		B, #0
		LD		C, A
		ADD		HL, BC
		LD		H, (HL)

		LD		A, D
		CP		E
		JR		C, graph_line_400


graph_line_300:						; 
		LD		C, D				; value = width
		LD		B, D				; count = width + 1
		INC		B
		LD		L, (IX)				; E     = *p_vram
graph_line_310:
		LD		A, L				; data |= bit
		OR		A, H
		LD		L, A

		RLC		H					; bit <<<= 1

		JP		NC, graph_line_320

		LD		(IX), L
		INC		IX
		LD		L, (IX)

graph_line_320:
		LD		A, C				; value -= height
		SUB		A, E

		JP		NC, graph_line_330

		ADD		A, D				; value += width

		LD		(IX), L
		ADD		IX, SP
		LD		L, (IX)

graph_line_330:
		LD		C, A

		DJNZ	graph_line_310

		LD		(IX), L
		LD		SP, IY
		POP		IX
		RET


graph_line_400:						; 
		LD		C, E				; value = height
		LD		B, E				; count = height + 1
		INC		B
graph_line_410:
		LD		A, (IX)				; data |= bit
		OR		A, H
		LD		(IX), A

		ADD		IX, SP

		LD		A, C				; value -= width
		SUB		A, D

		JP		NC, graph_line_430

		ADD		A, E

		RLC		H					; bit <<<= 1

		JP		NC, graph_line_430

		INC		IX

graph_line_430:
		LD		C, A

		DJNZ	graph_line_410

		LD		SP, IY
		POP		IX
		RET
	__endasm ;
}
