#include <stdint.h>
#include "graph.h"
#include "sc1602bs.h"

#define sn76489_tone1Frequency( freq )		{ m_sn76489 = (0x80 | (freq & 15)) ;  m_sn76489 = (freq >> 4) ; }
#define sn76489_tone1Attenuation( atte )	{ m_sn76489 = (0x90 | (atte)) ;                                 }
#define sn76489_tone2Frequency( freq )		{ m_sn76489 = (0xa0 | (freq & 15)) ;  m_sn76489 = (freq >> 4) ; }
#define sn76489_tone2Attenuation( atte )	{ m_sn76489 = (0xb0 | (atte)) ;                                 }
#define sn76489_tone3Frequency( freq )		{ m_sn76489 = (0xc0 | (freq & 15)) ;  m_sn76489 = (freq >> 4) ; }
#define sn76489_tone3Attenuation( atte )	{ m_sn76489 = (0xd0 | (atte)) ;                                 }
#define sn76489_noiseControl(fb, nf)		{ m_sn76489 = (0xe0 | ((fb)<<2)) | (nf) ;                       }
#define sn76489_noiseAttenuation( atte )	{ m_sn76489 = (0xf0 | (atte)) ;                                 }


	enum {
		LINE_COUNT					=	8							/* ` */
	} ;

	/* ʒu */
	typedef struct {
		uint8_t						x ;								/* W   */
		uint8_t						y ;								/* W   */
		int8_t						ax ;							/*  */
		int8_t						ay ;							/*  */
	} POSINFO ;

	/*  */
	typedef struct {
		uint8_t						sx ;							/* JnW */
		uint8_t						sy ;							/* JnW */
		uint8_t						ex ;							/* IW */
		uint8_t						ey ;							/* IW */
	} LINEINFO ;


static volatile uint8_t __at (0xf000)	m_led ;						/* LED                          */
static volatile uint8_t __at (0xf100)	m_7seg[4] ;					/* 7SEG                         */
static volatile uint8_t __at (0xf300)	m_sn76489 ;					/* SN76489 Rg[WX^ */

static const uint8_t				m_cgramData[][8]={
										{ 0x00,0x00,0x00,0x00,0x00,0x01,0x1e,0x00 },
										{ 0x00,0x00,0x00,0x01,0x06,0x18,0x00,0x00 },
										{ 0x00,0x03,0x0c,0x10,0x00,0x00,0x00,0x00 },
										{ 0x0f,0x10,0x00,0x00,0x00,0x00,0x00,0x00 },
										{ 0x1e,0x01,0x00,0x00,0x00,0x00,0x00,0x00 },
										{ 0x00,0x18,0x06,0x01,0x00,0x00,0x00,0x00 },
										{ 0x00,0x00,0x00,0x10,0x0c,0x03,0x00,0x00 },
										{ 0x00,0x00,0x00,0x00,0x00,0x10,0x0f,0x00 },
									} ;

static const uint8_t				m_pattern[]={
										0x10,0x02,0x20,0x04,0x40,0x08,0x20,0x01
									} ;

static const uint8_t				m_number[]={
										0x5f,0x0a,0x76,0x7a,0x2b,0x79,0x7d,0x1a,0x7f,0x7b
									} ;

static uint8_t						m_atte[4] ;
static POSINFO						m_posInfo[2] ;
static LINEINFO						m_lineInfo[LINE_COUNT] ;


/* -------------------------------------------------------------------------------- */
/*                                                                            */
/*                                                                                  */
/*   IN : Ȃ                                                                      */
/*   OUT: Ȃ                                                                      */
/* -------------------------------------------------------------------------------- */
static void init( void )
{
	LINEINFO				*p_lineInfo ;
	uint8_t					index ;

	/*  */
	m_atte[0]       = 15 ;
	m_atte[1]       = 15 ;
	m_atte[2]       = 15 ;
	m_atte[3]       = 15 ;
	m_posInfo[0].x  = 12 ;
	m_posInfo[0].y  = 24 ;
	m_posInfo[0].ax =  5 ;
	m_posInfo[0].ay = -3 ;
	m_posInfo[1].x  = 10 ;
	m_posInfo[1].y  = 52 ;
	m_posInfo[1].ax = -3 ;
	m_posInfo[1].ay =  4 ;

	p_lineInfo = &m_lineInfo[0] ;
	for(index=0; index<LINE_COUNT; index++,p_lineInfo++) {
		p_lineInfo->sx = m_posInfo[0].x ;
		p_lineInfo->sy = m_posInfo[0].y ;
		p_lineInfo->ex = m_posInfo[1].x ;
		p_lineInfo->ey = m_posInfo[1].y ;
	}

	/* SC1602BS  */
	sc1602bs_init() ;
	sc1602bs_clearDisplay() ;
	sc1602bs_functionSet(1, 0) ;
	sc1602bs_displayOnOfCursor(1, 0, 0) ;
	sc1602bs_entryModeSet(1, 1) ;

	/* CGRAM ݒ */
	sc1602bs_cgramAddressSet( 0 ) ;
	for(index=0; index<sizeof( m_cgramData ); index++) {
		sc1602bs_cgramDdramDataWrite( m_cgramData[0][index] ) ;
	}

	/* DDRAM ݒ */
	sc1602bs_ddramAddressSet( 0 ) ;
	for(index=0; index<10; index++) {
		sc1602bs_cgramDdramDataWrite( 0 ) ;
		sc1602bs_cgramDdramDataWrite( 1 ) ;
		sc1602bs_cgramDdramDataWrite( 2 ) ;
		sc1602bs_cgramDdramDataWrite( 3 ) ;
		sc1602bs_cgramDdramDataWrite( 4 ) ;
		sc1602bs_cgramDdramDataWrite( 5 ) ;
		sc1602bs_cgramDdramDataWrite( 6 ) ;
		sc1602bs_cgramDdramDataWrite( 7 ) ;
	}

	/* SN76489  */
	sn76489_tone1Frequency( 320 ) ;
	sn76489_tone2Frequency( 220 ) ;
	sn76489_tone3Frequency( 120 ) ;
	sn76489_noiseControl(1, 0) ;
}


/* -------------------------------------------------------------------------------- */
/*   ʒuXV                                                                 */
/*                                                                                  */
/*   IN : p_posInfo = ʒu                                                      */
/*   OUT: Ȃ                                                                      */
/* -------------------------------------------------------------------------------- */
static void updatePosInfo( POSINFO *p_posInfo )
{
	int16_t		pos ;

	/* W */
	pos = (uint16_t )p_posInfo->x + p_posInfo->ax ;
	if( pos < 0 ) {
		pos           = 0 ;
		m_atte[0]     = 0 ;
		p_posInfo->ax = -p_posInfo->ax ;
	} else
	if( pos > (SCREEN_WIDTH-1) ) {
		pos           = (SCREEN_WIDTH-1) ;
		m_atte[1]     = 0 ;
		p_posInfo->ax = -p_posInfo->ax ;
	}
	p_posInfo->x = (uint8_t )pos ;

	/* W */
	pos = (uint16_t )p_posInfo->y + p_posInfo->ay ;
	if( pos < 0 ) {
		pos           = 0 ;
		m_atte[2]     = 0 ;
		p_posInfo->ay = -p_posInfo->ay ;
	} else
	if( pos > (SCREEN_HEIGHT-1) ) {
		pos           = (SCREEN_HEIGHT-1) ;
		m_atte[3]     = 0 ;
		p_posInfo->ay = -p_posInfo->ay ;
	}
	p_posInfo->y = (uint8_t )pos ;
}


/* -------------------------------------------------------------------------------- */
/*   C                                                                     */
/*                                                                                  */
/*   IN : Ȃ                                                                      */
/*   OUT: Ȃ                                                                      */
/* -------------------------------------------------------------------------------- */
int main( void )
{
	LINEINFO	*p_lineInfo ;
	uint8_t		index, count ;

	/*  */
	init() ;

	count = 0 ;

	while( 1 ) {
		/* ʃVtg */
		sc1602bs_cursorDisplayShift(1, 0) ;

		/* ʒuXV */
		updatePosInfo( &m_posInfo[0] ) ;
		updatePosInfo( &m_posInfo[1] ) ;

		p_lineInfo     = &m_lineInfo[count%LINE_COUNT] ;
		p_lineInfo->sx = m_posInfo[0].x ;
		p_lineInfo->sy = m_posInfo[0].y ;
		p_lineInfo->ex = m_posInfo[1].x ;
		p_lineInfo->ey = m_posInfo[1].y ;

		/* zʃRs[ */
		graph_copy() ;

		/* zʃNA */
		graph_clear() ;

		/* ` */
		p_lineInfo = &m_lineInfo[0] ;
		for(index=0; index<LINE_COUNT; index++,p_lineInfo++) {
			graph_line(p_lineInfo->sx, p_lineInfo->sy, p_lineInfo->ex, p_lineInfo->ey) ;
		}

		/* XV */
		if( m_atte[0] != 15 ) {
			sn76489_tone1Attenuation( m_atte[0] ) ;
			m_atte[0]++ ;
		}

		if( m_atte[1] != 15 ) {
			sn76489_tone1Attenuation( m_atte[1] ) ;
			m_atte[1]++ ;
		}

		if( m_atte[2] != 15 ) {
			sn76489_tone2Attenuation( m_atte[2] ) ;
			m_atte[2]++ ;
		}

		if( m_atte[3] != 15 ) {
			sn76489_noiseAttenuation( m_atte[3] ) ;
			m_atte[3]++ ;
		}

		/* LED XV */
		m_led = 1 << (count & 7) ;

		/* 7SEG XV */
		m_7seg[0] = m_number[count/100%10] ;
		m_7seg[1] = m_number[count/10 %10] ;
		m_7seg[2] = m_number[count    %10] ;
		m_7seg[3] = m_pattern[count&7] ;

		count++ ;
	}
}
